-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

use work.sha3_jh_package.all;
use work.sha3_pkg.all;

entity jh_datapath_mem is	
	generic ( h : integer := 256);
	port (
		-- external
		clk : in std_logic;
		din  : in std_logic_vector(w-1 downto 0);	
		dout : out std_logic_vector(w-1 downto 0);	

		--fsm 1		
		wmin1, wmin2, wmin3, wmin4 : in std_logic;
		din_addr : in std_logic_vector(log2(b/fbw)-1 downto 0);
		ec, lc, li, ein : in std_logic;	
		zc0 : out std_logic;
		--fsm 2			  		
		round : in std_logic_vector(log2(64*b/fbw)-1 downto 0);
		r_addr1_1, r_addr1_2, r_addr2_1, r_addr2_2 : in std_logic_vector(log2(b/fbw)-1 downto 0);
		whp, wr1, wr2 : in std_logic;
		sf, srdp, srin, srout1, srout2 : in std_logic;			
		
		--fsm 3
		sdout : in std_logic_vector(h/256-1 downto 0);
		eout : in std_logic
	);				  
end jh_datapath_mem;


architecture struct of jh_datapath_mem is 
	--input		  
	signal min : std_logic_vector(fbw/2-1 downto 0);
	type inRAM_out is array (0 to 3) of std_logic_vector(fbw/8-1 downto 0);
	signal in_RAM_out : inRAM_out;
	signal wmin : std_logic_vector(3 downto 0);
	signal piso_in : std_logic_vector(fbw/8-1 downto 0);
	signal min_mem : std_logic_vector(fbw/2-1 downto 0);	

	-- round constant
	signal crd_out : std_logic_vector(fbw/4-1 downto 0);
	
	-- round		
	signal rd_out :  std_logic_vector(fbw-1 downto 0);
	
	signal g, dg : std_logic_vector(fbw-1 downto 0);
	signal rin, rout : std_logic_vector(fbw-1 downto 0);
	signal rin1, rin2, rout1, rout2 : std_logic_vector(fbw/2-1 downto 0);
	type rRAM_out is array (0 to 3) of std_logic_vector(fbw/2-1 downto 0);
	signal r_out : rRAM_out;  -- rout1_1, rout1_2, rout2_1, and rout2_2
	type rRAM_addr is array (0 to 3) of std_logic_vector(log2(b/fbw)-1 downto 0);
	signal r_addr : rRAM_addr;
	signal wr : std_logic_vector(1 downto 0);
	signal hp_out : std_logic_vector(fbw/8-1 downto 0);
	signal hp, hm : std_logic_vector(fbw-1 downto 0);
	signal hp_or_iv : std_logic_vector(fbw-1 downto 0);
	signal c : std_logic_vector(15 downto 0);
	
	type iv_rom is array (0 to 1024/fbw-1) of std_logic_vector(fbw-1 downto 0);
	function split_iv ( input : std_logic_vector; split_size : integer ) return iv_rom;
	constant iv : iv_rom := split_iv (get_iv( h ), fbw); 
	
	type cr_rom is array (0 to 64*b/fbw-1) of std_logic_vector(fbw/4-1 downto 0);
	function split_cr ( input : rc_type; split_size : integer ) return cr_rom;
	constant rc_cons_fv : cr_rom := split_cr(rc_cons, fbw);
	
	
	-- Functions to create IV ROM and CR ROM
	function split_iv ( input : std_logic_vector; split_size : integer ) return iv_rom is
		variable retval : iv_rom;
		variable bits : integer := split_size / 8;
	begin
		for i in 0 to (1024 / split_size) - 1 loop
			retval(i) := input(1024 - bits * i - 1 downto 1024 - bits * (i + 1)) & input(896 - bits * i - 1 downto 896 - bits * (i + 1)) & 
				input(768 - bits * i - 1 downto 768 - bits * (i + 1)) & input(640 - bits * i - 1 downto 640 - bits * (i + 1)) &
				input(512 - bits * i - 1 downto 512 - bits * (i + 1)) & input(384 - bits * i - 1 downto 384 - bits * (i + 1)) &
				input(256 - bits * i - 1 downto 256 - bits * (i + 1)) & input(128 - bits * i - 1 downto 128 - bits * (i + 1));
		end loop;
		return retval;
	end split_iv;
	function split_cr ( input : rc_type; split_size : integer ) return cr_rom is
		variable retval : cr_rom;
	begin
		for i in 0 to 43 loop
			for j in 0 to (1024 / split_size) -1 loop
				retval(i * (1024 / split_size) + j) := input(i)(256 - (split_size/4) * j - 1 downto 256 - (split_size/4) * (j + 1));
			end loop;
		end loop;
		return retval;
	end split_cr;

begin						
	---//////////////---
	-- block decounter																									   	
	decounter_gen : decountern generic map ( N => 16, sub => 1 ) port map ( clk => clk, rst => '0', load => lc, en => ec, input => din(log2mw+15 downto log2mw), output => c);	
	zc0 <= '1' when c = 0 else '0';																																										

	shfin_gen : piso generic map ( N => w, M=> fbw/8 ) 
						port map ( clk => clk, sel => li, en => ein, 
									input => din, output => piso_in );

	-- input
	wmin(0) <= wmin1;
	wmin(1) <= wmin2;
	wmin(2) <= wmin3;
	wmin(3) <= wmin4;
	in_RAM_gen : for i in 0 to 3 generate
		din_ram : ramn generic map ( M => fbw/8, N => b/fbw) 
						port map ( input => piso_in, 
						addr => din_addr, 
						write => wmin(i), 
						clk => clk, 
						output => in_RAM_out(i));
	end generate;
	min <= in_RAM_out(0) & in_RAM_out(1) & in_RAM_out(2) & in_RAM_out(3);
	min_RAM_gen : ramn generic map ( M => fbw/2, N => b/fbw) 
					port map ( input => min, 
								addr => r_addr1_1, 
								write => whp, 
								clk => clk, 
								output => min_mem);
	
	-- input to group
	hm <= ( min xor hp_or_iv(fbw-1 downto fbw/2) ) & hp_or_iv(fbw/2-1 downto 0);	
	hp_or_iv <= iv(to_integer(unsigned(din_addr))) when sf = '1' else hp;	
	
	-- group (rearrange them into correct order)		   
	g <= form_group( hm, fbw, fbw/4 );
	rin <= g when srdp = '1' else rd_out;

	--R ram
	rin1 <= rin(fbw-1 downto fbw/2) when srin = '1' else rin(fbw/2-1 downto 0);
	rin2 <= rin(fbw-1 downto fbw/2) when srin = '0' else rin(fbw/2-1 downto 0);
	r_addr(0) <= r_addr1_1;
	r_addr(1) <= r_addr2_1;
	r_addr(2) <= r_addr1_2;
	r_addr(3) <= r_addr2_2;
	wr(0) <= wr1;
	wr(1) <= wr2;
	r_RAM_gen : for i in 0 to 1 generate
		r_ram1 : ramn generic map ( M => fbw/2, N => b/fbw) 
						port map ( input => rin1, 
						addr => r_addr(2*i), 
						write => wr(i), 
						clk => clk, 
						output => r_out(2*i));
		r_ram2 : ramn generic map ( M => fbw/2, N => b/fbw) 
						port map ( input => rin2, 
						addr => r_addr(2*i+1), 
						write => wr(i), 
						clk => clk, 
						output => r_out(2*i+1));
	end generate;
	rout1 <= r_out(0) when srout1 = '1' else r_out(2);
	rout2 <= r_out(1) when srout1 = '1' else r_out(3);
	rout <= (rout1 & rout2) when srout2 = '1' else (rout2 & rout1);
		
	-- output to round function	 
	dg <= degroup( rout, fbw, fbw/4 );
 	hp <= dg(fbw-1 downto fbw/2) & ( min_mem xor dg(fbw/2-1 downto 0) );

	-- output
	h256 : if h=256 generate
		hp_out <= hp(fbw/4-1 downto fbw/8) when sdout = "0" else hp(fbw/8-1 downto 0);
	end generate;
	
	h512 : if h=512 generate
		hp_out <= hp(fbw/2-1 downto 3*fbw/8) when sdout = "00" else
					hp(3*fbw/8-1 downto fbw/4) when sdout = "01" else
					hp(fbw/4-1 downto fbw/8) when sdout = "10" else 
					hp(fbw/8-1 downto 0);
	end generate;	
		
	shfout_gen : sipo generic map ( M => fbw/8, N => w) 
						port map ( clk => clk, en => eout, 
									input => hp_out, output => dout );

	-- RC 
	crd_out <= rc_cons_fv(to_integer(unsigned(round)));
	-- ROUND
	rd_gen 	: entity work.jh_rd(struct) generic map ( bw => fbw, cw => fbw/4  ) 
										port map ( input => rout, cr => crd_out, output => rd_out );
  	 
	
end struct;		   
