-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================

library IEEE;
use ieee.std_logic_1164.all;			
use ieee.std_logic_unsigned.all; 
use ieee.std_logic_arith.all;
use work.sha3_jh_package.all;
use work.sha3_pkg.all;

entity jh_fsm1 is	   
	port (
	clk : in std_logic;
	rst : in std_logic;		
	
	-- datapath sigs
	zc0, final_segment	: in std_logic;
	lc, ec : out std_logic;
	li, ein : out std_logic;
	addr	: out std_logic_vector(fmw-3 downto 0);
	w1, w2, w3, w4	: out std_logic;
	
	-- Control communication
	load_next_block	: in std_logic;		 
	msg_ready : out std_logic;
	msg_end : out std_logic;
	
	-- FIFO communication
	src_ready : in std_logic;
	src_read   : out std_logic	
	);
end jh_fsm1;

architecture beh of jh_fsm1 is	
	-- counter
	signal ebc : std_logic;		 
	signal msg_count : std_logic_vector(fmw-1 downto 0);
	
	-- fsm sigs
	type state_type is ( reset, wait_for_header1, wait_for_header2, wait_for_block, load_block, write_block, ready );
	signal cstate, nstate : state_type; 			 
	
	--
	signal f, ef, cf : std_logic;
	signal done_load, done_write : std_logic;
begin
	-- final seg register
	sr_final_segment :  sr_reg 
	port map ( rst => rst, clk => clk, set => ef, clr => cf, output => f);
	
	-- fsm1 counter						
	block_counter_gen : countern generic map ( n => fmw ) port map (clk => clk, rst => rst, load => load_next_block, en => ebc , input => fmwzeros, output => msg_count);
	addr <= msg_count(fmw-3 downto 0);	
	
	-- state process
	cstate_proc : process ( clk )
	begin
		if rising_edge( clk ) then 
			if rst = '1' then
				cstate <= reset;
			else
				cstate <= nstate;
			end if;
		end if;
	end process;

	
	nstate_proc : process ( cstate, src_ready, load_next_block, done_write, done_load, zc0, final_segment, f )
	begin
		case cstate is
			when reset =>
				nstate <= wait_for_header1;
			when wait_for_header1 =>
				if ( src_ready = '1' ) then
					nstate <= wait_for_header1;
				elsif ( final_segment = '1' ) then
					nstate <= wait_for_header2;
				else 
					nstate <= wait_for_block;
				end if;
			when wait_for_header2 =>
				if (src_ready = '1') then
					nstate <= wait_for_header2;
				else
					nstate <= wait_for_block;
				end if;	 
			when wait_for_block =>
				if (src_ready = '1') then
					nstate <= wait_for_block;
				else
					nstate <= load_block;
				end if;	 				
			when load_block =>		
				nstate <= write_block;										 					
			when write_block =>
				if (done_write = '1') then
					if (done_load = '1') then
						nstate <= ready;
					elsif (src_ready='0') then
						nstate <= load_block;
					else
						nstate <= wait_for_block;
					end if;
				else
					nstate <= write_block;
				end if;
			when ready =>
				if (load_next_block = '0') then
					nstate <= ready;
				elsif (zc0= '1') then
					if (src_ready = '1') then
						nstate <= wait_for_header1;
					elsif ( final_segment = '1' ) then
						nstate <= wait_for_header2;
					else 
						nstate <= wait_for_block;
					end if;
				elsif (src_ready = '1') then
					nstate <= wait_for_block;
				else 
					nstate <= load_block;
				end if;
		end case;
	end process;
	
	-- fsm output
	w1 <= '1' when (msg_count(fmw-1 downto fmw-2) = "00" and cstate = write_block) else '0';
	w2 <= '1' when (msg_count(fmw-1 downto fmw-2) = "01" and cstate = write_block) else '0';
	w3 <= '1' when (msg_count(fmw-1 downto fmw-2) = "10" and cstate = write_block) else '0';
	w4 <= '1' when (msg_count(fmw-1 downto fmw-2) = "11" and cstate = write_block) else '0';
	
	src_read <= '1' when (((cstate = wait_for_header1 or cstate = wait_for_header2 or 
							(cstate = ready and load_next_block = '1' and zc0 = '1')) and 
							src_ready = '0') or cstate = load_block) else '0';
	
	done_write <= '1' when (msg_count(log2(w*8/fbw)-1 downto 0) = fmwones(log2(w*8/fbw)-1 downto 0)) else '0';
	done_load <= '1' when (msg_count(fmw-1 downto fmw-3) = "111") else '0';
	ebc <= '1' when (cstate = write_block) else '0';
	ein <= '1' when (cstate = write_block or cstate = load_block) else '0';
	li <= '1' when (cstate = load_block) else '0';
						 
    msg_ready <= '1' when (cstate = ready) else '0';
	msg_end <= '1' when (cstate = ready and zc0 = '1' and f = '1') else '0';
		
	ef <= '1' when ((cstate = wait_for_header1 or (cstate = ready and zc0 = '1' and load_next_block = '1')) and 
					src_ready = '0' and final_segment = '1')  else '0';	   
	cf <= '1' when (cstate = ready and load_next_block = '1' and zc0 = '1') else '0';
	
	lc <= '1' when (src_ready = '0' and (cstate = wait_for_header1 or 
					(cstate = ready and load_next_block = '1' and zc0 = '1'))) else '0';
		 			
	ec <= '1' when 	(cstate = write_block and done_write = '1' and done_load = '1') else '0';		
		
end beh;