-- =====================================================================
-- Copyright © 2010-2012 by Cryptographic Engineering Research Group (CERG),
-- ECE Department, George Mason University
-- Fairfax, VA, U.S.A.
-- =====================================================================
 
library ieee;
use ieee.std_logic_1164.all;	   
use ieee.std_logic_arith.all;
use work.sha3_pkg.all;
use work.sha3_jh_package.all;		  

entity jh_fsm3 is 
	generic ( 
		h : integer := 256
	);
	port (
		-- global
		clk : in std_logic;
		rst : in std_logic;
		
		-- datapath
		eout, srout1, srout2 : out std_logic;
		sdout : out std_logic_vector(h/256-1 downto 0);
		r_addr : out std_logic_vector(log2(b/fbw)-1 downto 0);

		-- control				   
			--fsm1 hand shake signals
		msg_end : in std_logic;
		hash_complete : out std_logic;
		
		-- FIFO communication
		dst_ready 				: in std_logic;
		dst_write				: out std_logic
		);							   
end jh_fsm3;

architecture beh of jh_fsm3 is	   

	type state_type is ( reset, idle, load_sipo, wait_for_dst, send_sipo );
	signal cstate, nstate : state_type;
	
	constant log2czeros : std_logic_vector(log2(fbw/2)-2+(h/256) downto 0) := (others => '0');
	constant log2cones : std_logic_vector(log2(8*w/fbw)-1 downto 0) := (others => '1');
	signal pc : std_logic_vector(log2(fbw/2)-2+(h/256) downto 0);
	signal lc, ec : std_logic;
	signal SIPO_loaded, hash_done : std_logic;
	
begin
	-- fsm3 counter
	proc_counter_gen : countern generic map ( N => log2(fbw/2)-1+(h/256) ) port map ( clk => clk, rst => rst, load => lc, en => ec, input => log2czeros, output => pc);
	
	-- state process
	cstate_proc : process ( clk )
	begin
		if rising_edge( clk ) then 
			if rst = '1' then
				cstate <= reset;
			else
				cstate <= nstate;
			end if;
		end if;
	end process;
	
	nstate_proc : process ( cstate, SIPO_loaded, msg_end, dst_ready, hash_done )
	begin
		case cstate is	 
			when reset =>
				nstate <= idle;
			when idle =>
				if msg_end = '1' then
					nstate <= load_sipo;
				else
					nstate <= idle;
				end if;
			when load_sipo => 
				if SIPO_loaded = '1' then
					if dst_ready = '0' then
						nstate <= send_sipo;
					else
						nstate <= wait_for_dst;
					end if;
				else
					nstate <= load_sipo;
				end if;
			when wait_for_dst =>
				if dst_ready = '0' then
					nstate <= send_sipo;
				else
					nstate <= wait_for_dst;
				end if;
			when send_sipo =>
				if hash_done = '1' then
					nstate <= idle;
				else
					nstate <= load_sipo;
				end if;
		end case;
	end process;
	
	---- internal logic
	SIPO_loaded <= '1' when (pc(log2(8*w/fbw)-1 downto 0) = log2cones) else '0'; 
	lc <= '1' when cstate = idle else '0';
	ec <= '1' when cstate = load_sipo else '0';
	
		
	---- output logic
	hash_complete <= '1' when (cstate = send_sipo and hash_done = '1') else '0';
	
	srout1 <= '1';
	srout2 <= not pc(log2(fbw/2)-2);
	
	sdout <= pc(log2(fbw/2)-2+h/256 downto log2(fbw/2)-1);
	hash_done <= '1' when (pc = log2czeros and cstate = send_sipo) else '0';

	
	r_addr <= pc(log2(b/fbw)-1 downto 0);
	
	eout <= '1' when (cstate = load_sipo) else '0';
	
	dst_write <= '1' when (cstate = send_sipo) else '0';
	
end beh;